/* Copyright or (C) or Copr. GET / ENST, Telecom-Paris, Ludovic Apvrille
 *
 * ludovic.apvrille AT enst.fr
 *
 * This software is a computer program whose purpose is to allow the
 * edition of TURTLE analysis, design and deployment diagrams, to
 * allow the generation of RT-LOTOS or Java code from this diagram,
 * and at last to allow the analysis of formal validation traces
 * obtained from external tools, e.g. RTL from LAAS-CNRS and CADP
 * from INRIA Rhone-Alpes.
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
package avatartranslator.tosysmlv2;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.function.BiConsumer;

import avatartranslator.*;
import avatartranslator.tosysmlv2.AvatarFromSysMLSyntax.*;
import java_cup.runtime.ComplexSymbolFactory;

public class AvatarFromSysML {
    private AvatarSpecification avSpec;
    private StxModel stxSpec;
    private ArrayList<AvatarTransition> transitionList;
    private HashMap<StxSignal, AvatarSignal> signalMap;
    private HashMap<StxBlock, AvatarBlock> blockMap;
    private HashMap<StxState, AvatarStateMachineElement> stateMap;
    private AvatarBlock getBlock(StxBlock _b) {
        AvatarBlock b = blockMap.get(_b);
        if (b == null) {
            b = new AvatarBlock(_b.getName(), avSpec, null);
            blockMap.put(_b, b);
        }
        return b;
    }
    private AvatarSignal getSignal(StxSignal _b) {
        AvatarSignal b = signalMap.get(_b);
        if (b == null) {
            if (_b.isInput())
                b = new AvatarSignal(_b.getName(), AvatarSignal.IN, null);
            else
                b = new AvatarSignal(_b.getName(), AvatarSignal.OUT, null);
            signalMap.put(_b, b);
        }
        return b;
    }
    private AvatarAttribute getAttributeByName(String _name, AvatarBlock _block) {
        List<AvatarAttribute> l = _block.getAttributes();
        int size = l.size();
        for (int i = 0; i < size; i++)
            if(l.get(i).getName().equals(_name)) return  l.get(i);
        return null;
    }
    private AvatarStateMachineElement getState(StxState _s, AvatarBlock _b) {
        AvatarStateMachineElement s = stateMap.get(_s);
        if (s == null) {
            switch(_s.getType()) {
                case AvatarFromSysMLSyntax.STXSTARTSTATE :
                    s = new AvatarStartState("StartState", null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXSTOPSTATE :
                    s = new AvatarStopState("StopState", null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXSTANDARDSTATE :
                    s = new AvatarState(_s.getName(), null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXRANDOMSTATE :
                    s = new AvatarRandom(_s.getName(), null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXCOUNTSTATE :
                    s = new AvatarQueryOnSignal(_s.getName(), getSignal(_s.getSignal()),
                            getAttributeByName(_s.getVariable(),_b), null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXSENDSTATE :
                case AvatarFromSysMLSyntax.STXRECEIVESTATE :
                    s = new AvatarActionOnSignal(_s.getName(), null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXPRESENDSTATE :
                case AvatarFromSysMLSyntax.STXPRERECEIVESTATE : break;
                case AvatarFromSysMLSyntax.STXSETTIMERSTATE :
                    s = new AvatarSetTimer(_s.getName(), null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXRESETTIMERSTATE :
                    s = new AvatarResetTimer(_s.getName(), null, _b);
                    break;
                case AvatarFromSysMLSyntax.STXEXPIRETIMERSTATE :
                    s = new AvatarExpireTimer(_s.getName(), null, _b);
            }
            if (s != null) stateMap.put(_s, s);
        }
        return s;
    }

    public void sysMLtoSpec(String _txt) {
        AvatarFromSysMLParser parser =
                new AvatarFromSysMLParser(new Avatar2SysMLLexer(new StringReader(_txt)),
                                          new ComplexSymbolFactory());
        stxSpec = parser.parseModel();
        buildDataTypes();
        buildBlocks();
    }
    // BUILDING DATATYPES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    private class BuildDataType implements BiConsumer<String, StxDataType> {
        public BuildDataType(){}
        public void accept(String n, StxDataType d) {
            AvatarDataType dataType = new AvatarDataType(d.getName());
            AvatarDataType.dataTypeMap.put(d.getName(), dataType);
            avSpec.getDataTypes().add(dataType);
            int nbFields = d.getSize();
            for (int i = 0; i < nbFields; i++) {
                String type =  d.getFieldType(i);
                if(type.equals("Integer"))
                    dataType.addField(d.getFieldName(i), AvatarDataType.INTEGER, null);
                else if(type.equals("Boolean"))
                    dataType.addField(d.getFieldName(i), AvatarDataType.BOOLEAN, null);
                else
                    dataType.addField(d.getFieldName(i), AvatarDataType.DATATYPE, type);
            }
        }
    }
    private void buildDataTypes(){
        List<AvatarDataType> dtList = avSpec.getDataTypes();
        AvatarDataType.dataTypeMap.clear();
        stxSpec.getDataTypeMap().forEach(new BuildDataType());
        int size = dtList.size();
        for(int i = 0; i < size; i++)
            AvatarDataType.finalize(dtList.get(i));
    }
    // BUILDING BLOCKS (but not statemachines) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    private class BuildBlock implements BiConsumer<String, StxBlock> {
        List<AvatarDataType> dataTypeList;
        public BuildBlock(){}
        public void accept(String n, StxBlock blk) {
            AvatarBlock theBlock = getBlock(blk);
            avSpec.addBlock(theBlock);
            // add Attributes
            int size = blk.getNbAttributes();
            for (int i = 0; i < size; i++) {
                StxAttribute a = blk.getAttribute(i);
                AvatarAttribute aa;
                if (a.getType().equals("Integer")) {
                    aa = new AvatarAttribute(a.getName(), AvatarType.INTEGER, theBlock, null);
                    aa.setInitialValue(a.getInit());
                    theBlock.addAttribute(aa);
                    aa.setAsConstant(false);
                }
                else if (a.getType().equals("Boolean")) {
                    aa = new AvatarAttribute(a.getName(), AvatarType.BOOLEAN, theBlock, null);
                    aa.setInitialValue(a.getInit());
                    theBlock.addAttribute(aa);
                    aa.setAsConstant(false);
                }
                else {
                    AvatarDataType adt = AvatarDataType.dataTypeMap.get(a.getType());
                    int tsize = adt.getFullSize();
                    for(int j = 0; j < tsize; j++) {
                        aa = new AvatarAttribute(
                                a.getName() + "__" + adt.getFieldString(j),
                                (adt.getFieldStringType(j) == AvatarDataType.BOOLEAN ? AvatarType.BOOLEAN : AvatarType.INTEGER),
                                theBlock, null);
                        theBlock.addAttribute(aa);
                        aa.setAsConstant(false);
                        if (j == 0) aa.setDataType(adt);
                    }
                }
            }
            // add Constants
            size = blk.getNbConstants();
            for (int i = 0; i < size; i++) {
                StxAttribute a = blk.getConstant(i);
                AvatarAttribute aa;
                if (a.getType().equals("Integer")) {
                    aa = new AvatarAttribute(a.getName(), AvatarType.INTEGER, theBlock, null);
                    aa.setInitialValue(a.getInit());
                    theBlock.addAttribute(aa);
                    aa.setAsConstant(true);
                }
                else if (a.getType().equals("Boolean")) {
                    aa = new AvatarAttribute(a.getName(), AvatarType.BOOLEAN, theBlock, null);
                    aa.setInitialValue(a.getInit());
                    theBlock.addAttribute(aa);
                    aa.setAsConstant(true);
                }
                else {
                    AvatarDataType adt = AvatarDataType.dataTypeMap.get(a.getType());
                    int tsize = adt.getFullSize();
                    for(int j = 0; j < tsize; j++) {
                        aa = new AvatarAttribute(
                                a.getName() + "__" + adt.getFieldString(j),
                                (adt.getFieldStringType(j) == AvatarDataType.BOOLEAN ? AvatarType.BOOLEAN : AvatarType.INTEGER),
                                theBlock, null);
                        theBlock.addAttribute(aa);
                        aa.setAsConstant(true);
                        if (j == 0) aa.setDataType(adt);
                    }
                }
            }
            // add Timers
            size = blk.getNbTimers();
            for (int i = 0; i < size; i++) {
                AvatarAttribute aa = new AvatarAttribute(blk.getTimer(i).getName(), AvatarType.TIMER, theBlock, null);
                theBlock.addAttribute(aa);
                aa.setAsConstant(false);
            }
            // add Methods
            size = blk.getNbMethods();
            for (int i = 0; i < size; i++) {
                StxMethod sm = blk.getMethod(i);
                AvatarMethod am = new AvatarMethod(sm.getName(),null);
                theBlock.addMethod(am);
                buildProfile(sm, am, theBlock);
                String returnType = sm.getReturnType();
                if (returnType == null) continue;
                if (returnType.equals("Integer"))
                    am.addReturnParameter(new AvatarAttribute("return__0", AvatarType.INTEGER, theBlock, null));
                else if (returnType.equals("Boolean"))
                    am.addReturnParameter(new AvatarAttribute("return__0", AvatarType.BOOLEAN, theBlock, null));
                else {
                    AvatarDataType adt = AvatarDataType.dataTypeMap.get(returnType);
                    int nbFields = adt.getNbFields();
                    for (int j = 0; j < nbFields; j++) {
                        AvatarType type = (adt.getFieldStringType(j) == AvatarDataType.INTEGER ? AvatarType.INTEGER : AvatarType.BOOLEAN);
                        am.addReturnParameter(new AvatarAttribute("return__" + j, type, theBlock, null));
                    }
                }
            }
            // add Signals (build profile delayed)
            size = blk.getNbSignals();
            for (int i = 0; i < size; i++) {
                StxSignal ss = blk.getSignal(i);
                AvatarSignal as = getSignal(ss);
                theBlock.addSignal(as);
            }
            // add states
            StxState[] states = blk.getStates();
            AvatarStateMachine asm = theBlock.getStateMachine();
            AvatarStateMachineElement theSourceState;
            size = states.length;
            for (int i = 0; i < size; i++) {
                byte stateType = states[i].getType();
                if (stateType != AvatarFromSysMLSyntax.STXPRESENDSTATE && stateType != AvatarFromSysMLSyntax.STXPRERECEIVESTATE) {
                    theSourceState = getState(states[i], theBlock);
                    asm.addElement(theSourceState);
                    if (stateType == AvatarFromSysMLSyntax.STXRANDOMSTATE) {
                        ((AvatarRandom) theSourceState).setValues(states[i].getMinValue(), states[i].getMaxValue());
                        ((AvatarRandom) theSourceState).setVariable(states[i].getVariable());
                    }
                // add transitions
                    List<StxTransition> transitions = states[i].getTransitions();
                    int tsize = transitions.size();
                    for (int j = 0; j < tsize; j++) {
                        AvatarTransition theTransition = new AvatarTransition(theBlock,"", null);
                        transitionList.add(theTransition);
                        theSourceState.addNext(theTransition);
                        theTransition.setGuard(transitions.get(j).getGuard());
                        theTransition.setDelays(transitions.get(j).getMinDelay(), transitions.get(j).getMaxDelay());
                        setDistributionLaw(theTransition, transitions.get(j).getDelayDistributionLaw(),transitions.get(j).getDelayExtra());
                    }
                }
            }
        }
    }
    private void buildBlocks(){
        stxSpec.getBlockMap().forEach(new BuildBlock());
    }
    private void setDistributionLaw(AvatarTransition _t, String distributionLaw, HashMap<String,String> delayExtra) {
        int law = -1;
        int size;
        if (distributionLaw == null || distributionLaw.equals("")) law = AvatarTransition.DELAY_UNIFORM_LAW;
        else {
            size = _t.DISTRIBUTION_LAWS.length;
            for (int i = 0; i < size; i++) {
                if (_t.DISTRIBUTION_LAWS[i].equals(distributionLaw)) {
                    law = i;
                    break;
                }
            }
            if (law == -1) law = AvatarTransition.DELAY_UNIFORM_LAW;
        }
        String extra1 = _t.LABELS_OF_EXTRA_ATTRIBUTES_1[law];
        String extra2 = _t.LABELS_OF_EXTRA_ATTRIBUTES_2[law];
        size = _t.NB_OF_EXTRA_ATTRIBUTES[law];
        if (size > 0) {
            extra1 = delayExtra.get(extra1);
            if (size > 1) {
                extra2 = delayExtra.get(extra2);
            }
        }
        _t.setDistributionLaw(law, extra1, extra2);
    }
    private void buildProfile(StxStructure ss, AvatarMethod am, AvatarBlock b) {
        AvatarAttribute aa;
        int size = ss.getSize();
        for (int i = 0; i < size; i++) {
            String type = ss.getFieldType(i);
            if (type.equals("Integer")){
                aa = new AvatarAttribute(ss.getFieldName(i), AvatarType.INTEGER, b, null);
                am.addParameter(aa);
                aa.setAsConstant(false);
            } else if (type.equals("Boolean")) {
                aa = new AvatarAttribute(ss.getFieldName(i), AvatarType.BOOLEAN, b, null);
                am.addParameter(aa);
                aa.setAsConstant(false);
            } else {
                AvatarDataType adt = AvatarDataType.dataTypeMap.get(type);
                int tsize = adt.getFullSize();
                for(int j = 0; j < tsize; j++) {
                    aa = new AvatarAttribute(
                            ss.getFieldName(i) + "__" + adt.getFieldString(j),
                            (adt.getFieldStringType(j) == AvatarDataType.BOOLEAN ? AvatarType.BOOLEAN : AvatarType.INTEGER),
                            b, null);
                    am.addParameter(aa);
                    aa.setAsConstant(false);
                    if (j == 0) aa.setDataType(adt);
                }
            }
        }
    }
    private String getStxAttributeType(String name, StxBlock blk){
        int size = blk.getNbAttributes();
        for (int i = 0; i < size; i++) {
            if (blk.getAttribute(i).getName().equals(name)) return blk.getAttribute(i).getType();
        }
        return null;
    }
    private String getStxPathType(String s, StxBlock b) {
        String[] path = s.split("__");
        String type = getStxAttributeType(path[0], b);
        int size = path.length;
        for (int i = 1; i < size; i++) {
            AvatarDataType adt = AvatarDataType.dataTypeMap.get(type);
            if (adt == null) return null;
            int nbFields = adt.getNbFields();
            int j;
            for (j = 0; j < nbFields; j++)
                if (adt.getFieldName(j).equals(path[i])) break;
            if (j == nbFields) return null;
            int adtType = adt.getFieldType(j);
            if (adtType == AvatarDataType.INTEGER)
                type = "Integer";
            else if (adtType == AvatarDataType.BOOLEAN)
                type = "Boolean";
            else
                type = adt.getDataTypeName(j);
        }
        return type;
    }
    private String extendPath(String path, StxBlock b) {
        String type = getStxPathType(path, b);
        if (type.equals("Integer") || type.equals("Boolean")) return path;
        AvatarDataType adt = AvatarDataType.dataTypeMap.get(type);
        int size = adt.getFullSize() - 1;
        if (size == -1) return path;
        StringBuffer res = new StringBuffer();
        int i;
        for (i = 0; i < size; i++)
            res.append(path + "__" + adt.getFieldString(i) + ", ");
        res.append(path + "__" + adt.getFieldString(i));
        return res.toString();
    }
    private String extendIdent (String path, StxBlock b) {
        String result = extendPath(path, b);
        return (result.indexOf(',') == -1 ? result : "(" + result + ")");
    }
     private String extendCall (StxCall call, StxBlock b) {
        StringBuffer result = new StringBuffer();
        int size = call.getNbIdents();
        int i;
        for (i =0 ; i < size; i++) {
            result.append(call.getSegment(i));
            result.append(extendPath(call.getIdent(i), b));
        }
        result.append(call.getSegment(i));
        return result.toString();
     }
}
